/*
   +----------------------------------------------------------------------+
   | HipHop for PHP                                                       |
   +----------------------------------------------------------------------+
   | Copyright (c) 2010- Facebook, Inc. (http://www.facebook.com)         |
   +----------------------------------------------------------------------+
   | This source file is subject to version 3.01 of the PHP license,      |
   | that is bundled with this package in the file LICENSE, and is        |
   | available through the world-wide-web at the following url:           |
   | http://www.php.net/license/3_01.txt                                  |
   | If you did not receive a copy of the PHP license and are unable to   |
   | obtain it through the world-wide-web, please send a note to          |
   | license@php.net so we can mail you a copy immediately.               |
   +----------------------------------------------------------------------+
*/
#include "runtime/vm/translator/hopt/hhbctranslator.h"

#include "util/trace.h"
#include "runtime/ext/ext_continuation.h"
#include "runtime/vm/translator/translator-runtime.h"
#include "runtime/vm/translator/translator-x64.h"
#include "runtime/vm/stats.h"
#include "runtime/vm/unit.h"
#include "runtime/vm/runtime.h"
#include "runtime/vm/translator/hopt/irfactory.h"


namespace HPHP {
namespace VM {
namespace JIT {

TRACE_SET_MOD(hhir);

ArrayData* HhbcTranslator::lookupArrayId(int arrId) {
  return getCurUnit()->lookupArrayId(arrId);
}

StringData* HhbcTranslator::lookupStringId(int strId) {
  return getCurUnit()->lookupLitstrId(strId);
}

Func* HhbcTranslator::lookupFuncId(int funcId) {
  return getCurUnit()->lookupFuncId(funcId);
}

PreClass* HhbcTranslator::lookupPreClassId(int preClassId) {
  return getCurUnit()->lookupPreClassId(preClassId);
}

const NamedEntityPair& HhbcTranslator::lookupNamedEntityPairId(int id) {
  return getCurUnit()->lookupNamedEntityPairId(id);
}

SSATmp* HhbcTranslator::push(SSATmp* tmp) {
  assert(tmp);
  m_evalStack.push(tmp);
  return tmp;
}

void HhbcTranslator::refineType(SSATmp* tmp, Type type) {
  // If type is more refined than tmp's type, reset tmp's type to type
  IRInstruction* inst = tmp->getInstruction();
  if (type.strictSubtypeOf(tmp->getType())) {
    // If tmp is incref or move, then chase down its src
    Opcode opc = inst->getOpcode();
    if (opc == Mov || opc == IncRef) {
      refineType(inst->getSrc(0), type);
      tmp->setType(outputType(inst));
    } else if (tmp->getType().isNull() && type.isNull()) {
      // Refining Null to Uninit or InitNull is supported
      tmp->setType(type);
    } else {
      // At this point, we have no business refining the type of any
      // instructions other than the following, which all control
      // their destination type via a type parameter.
      //
      // FIXME: I think most of these shouldn't be possible still
      // (except LdStack?).
      //
      // XXX These are possible once we remove the inferred/predicted
      // type from emitCGetProp etc in HhbcTranslator. We need to
      // delete label on these instructions if this is due to an
      // assertType and also handled LdClsCns.
      // TODO(#2035446): fix this for LdClsCns
      assert(opc == LdLoc || opc == LdStack  ||
             opc == LdMem || opc == LdProp ||
             opc == LdRef);
      inst->setTypeParam(type);
      tmp->setType(type);
      assert(outputType(inst) == type);
    }
  }
}

SSATmp* HhbcTranslator::pop(Type type) {
  SSATmp* opnd = m_evalStack.pop();
  if (opnd == nullptr) {
    uint32_t stackOff = m_stackDeficit;
    m_stackDeficit++;
    return m_tb->genLdStack(stackOff, type);
  }

  // Refine the type of the temp given the information we have from
  // `type'.  This case can occur if we did an extendStack() and
  // didn't know the type of the intermediate values yet (see below).
  refineType(opnd, type);
  return opnd;
}

void HhbcTranslator::discard(unsigned n) {
  for (unsigned i = 0; i < n; ++i) {
    pop(Type::Gen);
  }
}

// type is the type expected on the stack.
void HhbcTranslator::popDecRef(Type type) {
  if (SSATmp* src = m_evalStack.pop()) {
    m_tb->genDecRef(src);
    return;
  }

  uint32_t stackOff = m_stackDeficit;
  m_tb->genDecRefStack(type, stackOff);
  m_stackDeficit++;
}

// We don't know what type description to expect for the stack
// locations before index, so we use a generic type when popping the
// intermediate values.  If it ends up creating a new LdStack,
// refineType during a later pop() or top() will fix up the type to
// the known type.
void HhbcTranslator::extendStack(uint32_t index, Type type) {
  if (index == 0) {
    push(pop(type));
    return;
  }

  SSATmp* tmp = pop(Type::Gen | Type::Cls);
  extendStack(index - 1, type);
  push(tmp);
}

SSATmp* HhbcTranslator::top(Type type, uint32_t index) {
  SSATmp* tmp = m_evalStack.top(index);
  if (!tmp) {
    extendStack(index, type);
    tmp = m_evalStack.top(index);
  }
  assert(tmp);
  refineType(tmp, type);
  return tmp;
}

void HhbcTranslator::replace(uint32_t index, SSATmp* tmp) {
  m_evalStack.replace(index, tmp);
}

void HhbcTranslator::setBcOff(Offset newOff, bool lastBcOff) {
  if (newOff != m_bcOff || m_bcOff == m_startBcOff) {
    m_bcOff = newOff;

    MarkerData marker;
    marker.bcOff     = m_bcOff;
    marker.func      = getCurFunc();
    marker.stackOff  = m_tb->getSpOffset() +
                         m_evalStack.numCells() - m_stackDeficit;
    m_tb->gen(Marker, &marker);
  }
  m_lastBcOff = lastBcOff;
}

void HhbcTranslator::emitPrint() {
  Type type = topC()->getType();
  if (type.subtypeOf(Type::Int | Type::Bool | Type::Null | Type::Str)) {
    Opcode op;
    if (type.isString()) {
      op = PrintStr;
    } else if (type.subtypeOf(Type::Int)) {
      op = PrintInt;
    } else if (type.subtypeOf(Type::Bool)) {
      op = PrintBool;
    } else {
      assert(type.isNull());
      op = Nop;
    }
    // the print helpers decref their arg, so don't decref pop'ed value
    if (op != Nop) m_tb->gen(op, popC());
    push(m_tb->genDefConst<int64_t>(1));
  } else {
    emitInterpOne(Type::Int, 1);
  }
}

void HhbcTranslator::emitUnboxRAux() {
  Block* exit = getExitTrace()->front();
  SSATmp* srcBox = popR();
  SSATmp* unboxed = m_tb->gen(Unbox, exit, srcBox);
  if (unboxed == srcBox) {
    // If the Unbox ended up being a noop, don't bother refcounting
    push(unboxed);
  } else {
    pushIncRef(unboxed);
    m_tb->genDecRef(srcBox);
  }
}

void HhbcTranslator::emitUnboxR() {
  TRACE(3, "%u: UnboxR\n", m_bcOff);
  emitUnboxRAux();
}

void HhbcTranslator::emitThis() {
  TRACE(3, "%u: This\n", m_bcOff);
  if (!getCurClass()) {
    emitInterpOne(Type::Obj, 0); // will throw a fatal
    return;
  }
  pushIncRef(m_tb->genLdThis(getExitSlowTrace()));
}

void HhbcTranslator::emitCheckThis() {
  TRACE(3, "%u: CheckThis\n", m_bcOff);
  if (!getCurClass()) {
    emitInterpOne(Type::None, 0); // will throw a fatal
    return;
  }
  m_tb->genLdThis(getExitSlowTrace());
}

void HhbcTranslator::emitBareThis(int notice) {
  TRACE(3, "%u: BareThis %d\n", m_bcOff, notice);
  // We just exit the trace in the case $this is null. Before exiting
  // the trace, we could also push null onto the stack and raise a
  // notice if the notice argument is set. By exiting the trace when
  // $this is null, we can be sure in the rest of the trace that we
  // have the this object on top of the stack, and we can eliminate
  // further null checks of this.
  if (!getCurClass()) {
    emitInterpOne(Type::InitNull, 0); // will raise notice and push null
    return;
  }
  pushIncRef(m_tb->genLdThis(getExitSlowTrace()));
}

void HhbcTranslator::emitArray(int arrayId) {
  TRACE(3, "%u: Array %d\n", m_bcOff, arrayId);
  ArrayData* ad = lookupArrayId(arrayId);
  push(m_tb->genDefConst(ad));
}

void HhbcTranslator::emitNewArray(int capacity) {
  TRACE(3, "%u: NewArray %d\n", m_bcOff, capacity);
  if (capacity == 0) {
    ArrayData* ad = HphpArray::GetStaticEmptyArray();
    push(m_tb->genDefConst(ad));
  } else {
    push(m_tb->genNewArray(capacity));
  }
}

void HhbcTranslator::emitNewTuple(int numArgs) {
  // The new_tuple helper function needs array values passed to it
  // via the stack.  We use spillStack() to flush the eval stack and
  // obtain a pointer to the topmost item; if over-flushing becomes
  // a problem then we should refactor the NewTuple opcode to take
  // its values directly as SSA operands.
  TRACE(3, "%u: NewTuple %d\n", m_bcOff, numArgs);
  SSATmp* sp = spillStack();
  for (int i = 0; i < numArgs; i++) popC();
  push(m_tb->genNewTuple(numArgs, sp));
}

void HhbcTranslator::emitArrayAdd() {
  TRACE(3, "%u: ArrayAdd\n", m_bcOff);
  Type type1 = topC(0)->getType();
  Type type2 = topC(1)->getType();
  if (!type1.isArray() || !type2.isArray()) {
    // This happens when we have a prior spillstack that optimizes away
    // its spilled values because they were already on the stack. This
    // prevents us from getting to type of the SSATmps popped from the
    // eval stack. Most likely we had an interpone before this
    // instruction.
    emitInterpOne(Type::Arr, 2);
    return;
  }
  SSATmp* tr = popC();
  SSATmp* tl = popC();
  // the ArrrayAdd helper decrefs its args, so don't decref pop'ed values
  // TODO task 1805916: verify that ArrayAdd increfs its result
  push(m_tb->gen(ArrayAdd, tl, tr));
}

void HhbcTranslator::emitAddElemC() {
  TRACE(3, "%u: AddElemC\n", m_bcOff);
  SSATmp* val = popC();
  SSATmp* key = popC();
  SSATmp* arr = popC();
  // the AddElem* instructions decrefs their args, so don't decref
  // pop'ed values. TODO task 1805916: verify that AddElem* increfs
  // their result
  auto kt = key->getType();
  Opcode op;
  if (kt.subtypeOf(Type::Int)) {
    op = AddElemIntKey;
  } else if (kt.isString()) {
    op = AddElemStrKey;
  } else {
    PUNT(AddElem-NonIntNonStr);
  }

  push(m_tb->gen(op, arr, key, val));
}

void HhbcTranslator::emitAddNewElemC() {
  TRACE(3, "%u: AddNewElemC\n", m_bcOff);
  SSATmp* val = popC();
  SSATmp* arr = popC();
  // the AddNewElem helper decrefs its args, so don't decref pop'ed values
  // TODO task 1805916: verify that NewElem increfs its result
  push(m_tb->gen(AddNewElem, arr, val));
}

void HhbcTranslator::emitNewCol(int type, int numElems) {
  emitInterpOneOrPunt(Type::Obj);
}

void HhbcTranslator::emitColAddElemC() {
  emitInterpOneOrPunt(Type::Obj, 3);
}

void HhbcTranslator::emitColAddNewElemC() {
  emitInterpOneOrPunt(Type::Obj, 2);
}

void HhbcTranslator::emitCns(uint32_t id) {
  emitInterpOneOrPunt(Type::Cell);
}

void HhbcTranslator::emitDefCns(uint32_t id) {
  TRACE(3, "%u: DefCns %d\n", m_bcOff, id);
  StringData* name = lookupStringId(id);
  SSATmp* val = popC();
  push(m_tb->genDefCns(name, val));
}

void HhbcTranslator::emitConcat() {
  SSATmp* tr = popC();
  SSATmp* tl = popC();
  // the concat helpers decref their args, so don't decref pop'ed values
  push(m_tb->genConcat(tl, tr));
}

void HhbcTranslator::emitDefCls(int cid, Offset after) {
//  m_tb->genDefCls(lookupPreClassId(cid), getCurUnit()->at(after));
  emitInterpOneOrPunt(Type::None);
}

void HhbcTranslator::emitDefFunc(int fid) {
//  m_tb->genDefFunc(lookupFuncId(fid));
  emitInterpOneOrPunt(Type::None);
}

void HhbcTranslator::emitLateBoundCls() {
  Class* clss = getCurClass();
  if (!clss) {
    // no static context class, so this will raise an error
    emitInterpOne(Type::Cls);
    return;
  }
  push(m_tb->gen(LdClsCtx, m_tb->genLdCtx(getCurFunc())));
}

void HhbcTranslator::emitSelf() {
  Class* clss = getCurClass();
  if (clss == NULL) {
    emitInterpOne(Type::Cls);
  } else {
    push(m_tb->genDefConst(clss));
  }
}

void HhbcTranslator::emitParent() {
  const Class* clss = getCurClass()->parent();
  if (clss == NULL || clss->parent() == NULL) {
    emitInterpOne(Type::Cls);
  } else {
    push(m_tb->genDefConst(clss->parent()));
  }
}

void HhbcTranslator::emitString(int strId) {
  TRACE(3, "%u: String %d\n", m_bcOff, strId);
  push(m_tb->genDefConst(lookupStringId(strId)));
}

void HhbcTranslator::emitInt(int64_t val) {
  TRACE(3, "%u: Int %" PRId64 "\n", m_bcOff, val);
  push(m_tb->genDefConst(val));
}

void HhbcTranslator::emitDouble(double val) {
  TRACE(3, "%u: Double %f\n", m_bcOff, val);
  push(m_tb->genDefConst(val));
}

void HhbcTranslator::emitNullUninit() {
  TRACE(3, "%u: NullUninit\n", m_bcOff);
  push(m_tb->genDefUninit());
}

void HhbcTranslator::emitNull() {
  TRACE(3, "%u: Null\n", m_bcOff);
  push(m_tb->genDefInitNull());
}

void HhbcTranslator::emitTrue() {
  TRACE(3, "%u: True\n", m_bcOff);
  push(m_tb->genDefConst(true));
}

void HhbcTranslator::emitFalse() {
  TRACE(3, "%u: False\n", m_bcOff);
  push(m_tb->genDefConst(false));
}

void HhbcTranslator::emitUninitLoc(uint32_t id) {
  TRACE(3, "%u: UninitLoc\n", m_bcOff);
  m_tb->genInitLoc(id, m_tb->genDefUninit());
}

void HhbcTranslator::emitInitThisLoc(int32_t id) {
  TRACE(3, "%u: InitThisLoc %d\n", m_bcOff, id);
  if (!getCurClass()) {
    // Do nothing if this is null
    return;
  }
  SSATmp* tmpThis = m_tb->genLdThis(getExitSlowTrace());
  m_tb->genInitLoc(id, m_tb->genIncRef(tmpThis));
}

void HhbcTranslator::emitCGetL(int32_t id) {
  TRACE(3, "%u: CGetL %d\n", m_bcOff, id);
  Trace* exitTrace = getExitTrace();
  pushIncRef(emitLdLocWarn(id, exitTrace));
}

void HhbcTranslator::emitCGetL2(int32_t id) {
  TRACE(3, "%u: CGetL2 %d\n", m_bcOff, id);
  Trace* exitTrace = getExitTrace();
  SSATmp* oldTop = pop(Type::Gen);
  pushIncRef(emitLdLocWarn(id, exitTrace));
  push(oldTop);
}

void HhbcTranslator::emitVGetL(int32_t id) {
  TRACE(3, "%u: VGetL %d\n", m_bcOff, id);
  pushIncRef(m_tb->genBoxLoc(id));
}

void HhbcTranslator::emitUnsetN() {
  // No reason to punt, translator-x64 does emitInterpOne as well
  emitInterpOne(Type::None, 1);
}

void HhbcTranslator::emitUnsetG(const StringData* gblName) {
  // No reason to punt, translator-x64 does emitInterpOne as well
  emitInterpOne(Type::None, 1);
}

void HhbcTranslator::emitUnsetL(int32_t id) {
  TRACE(3, "%u: UnsetL %d\n", m_bcOff, id);
  m_tb->genBindLoc(id, m_tb->genDefUninit());
}

void HhbcTranslator::emitBindL(int32_t id) {
  TRACE(3, "%u: BindL %d\n", m_bcOff, id);
  SSATmp* src = popV();
  if (m_curFunc->isPseudoMain()) {
    // in pseudo mains, the value of locals could change in functions
    // called explicitly (or implicitly via exceptions or destructors)
    // so we need to incref eagerly in case one of these functions
    // changes the value of our local and makes src dead.
    pushIncRef(src);
  }
  m_tb->genBindLoc(id, src);
  if (!m_curFunc->isPseudoMain()) {
    pushIncRef(src);
  }
}

void HhbcTranslator::emitSetL(int32_t id) {
  TRACE(3, "%u: SetL %d\n", m_bcOff, id);
  Trace* exitTrace = getExitTrace();
  SSATmp* src = popC();
  // Note we can't use the same trick as emitBindL in which we
  // move the incref to after the store because the stored location
  // might contain a ref, which could then be modified by the decref
  // inserted after the stloc
  push(m_tb->genStLoc(id, src, true, true, exitTrace));
}

void HhbcTranslator::emitIncDecL(bool pre, bool inc, uint32_t id) {
  // Handle only integer inc/dec for now
  Trace* exitTrace = getExitSlowTrace();
  SSATmp* src = m_tb->genLdLocAsCell(id, exitTrace);
  SSATmp* res = emitIncDec(pre, inc, src);
  m_tb->genStLoc(id, res, false, false, nullptr);
}

// only handles integer or double inc/dec
SSATmp* HhbcTranslator::emitIncDec(bool pre, bool inc, SSATmp* src) {
  assert(src->isA(Type::Int) || src->isA(Type::Dbl));
  SSATmp* one = src->isA(Type::Int) ? m_tb->genDefConst(1)
                                    : m_tb->genDefConst(1.0);
  SSATmp* res = inc ? m_tb->genAdd(src, one) : m_tb->genSub(src, one);
  // no incref necessary on push since result is an int
  push(pre ? res : src);
  return res;
}

void HhbcTranslator::emitIncDecMem(bool pre,
                                   bool inc,
                                   SSATmp* propAddr,
                                   Trace* exitTrace) {
  // Handle only integer inc/dec for now
  SSATmp* src = m_tb->genLdMem(propAddr, Type::Int, exitTrace);
  // do the add and store back
  SSATmp* res = emitIncDec(pre, inc, src);
  // don't gen a dec ref or type store
  m_tb->genStMem(propAddr, res, false);
}

static bool isSupportedBinaryArith(Opcode opc,
                                   Type t1,
                                   Type t2) {
  switch (opc) {
    // Opcodes supporting FP
    case OpAdd:
    case OpSub:
    case OpMul: return (t1.subtypeOf(Type::Int | Type::Bool | Type::Dbl) &&
                        t2.subtypeOf(Type::Int | Type::Bool | Type::Dbl));

    default:    return (t1.subtypeOf(Type::Int | Type::Bool) &&
                        t2.subtypeOf(Type::Int | Type::Bool));
  }
}

void HhbcTranslator::emitSetOpL(Opcode subOpc, uint32_t id) {
  TRACE(3, "%u: SetOpL %d\n", m_bcOff, id);
  Trace* exitTrace = getExitTrace();
  SSATmp* loc = emitLdLocWarn(id, exitTrace);
  if (subOpc == Concat) {
    // The concat helpers decref their args, so don't decref pop'ed values
    // and don't decref the old value held in the local. The concat helpers
    // also incref their results, which will be consumed by the stloc. We
    // need an extra incref for the push onto the stack.
    SSATmp* val = popC();
    SSATmp* result = m_tb->genConcat(loc, val);
    pushIncRef(m_tb->genStLoc(id, result, false, true, exitTrace));
  } else if (isSupportedBinaryArith(subOpc,
                                    loc->getType(),
                                    topC()->getType())) {
    SSATmp* val = popC();
    Type resultType = Type::binArithResultType(loc->getType(),
                                               val->getType());
    SSATmp* result = m_tb->gen(subOpc, resultType, loc, val);
    push(m_tb->genStLoc(id, result, true, true, exitTrace));
  } else {
    PUNT(SetOpL);
  }
}

void HhbcTranslator::emitClassExists(const StringData* clsName) {
  emitInterpOneOrPunt(Type::Bool, 2);
}

void HhbcTranslator::emitInterfaceExists(const StringData* ifaceName) {
  emitClassExists(ifaceName);
}

void HhbcTranslator::emitTraitExists(const StringData* traitName) {
  emitClassExists(traitName);
}

void HhbcTranslator::emitStaticLocInit(uint32_t varId, uint32_t listStrId) {
  emitInterpOneOrPunt(Type::None, 1);
}

void HhbcTranslator::emitReqDoc(const StringData* name) {
  PUNT(ReqDoc);
// Can't interp one req instructions because their interp one
// function changes the pc, sp, and fp.
//  spillStack();
//  popC();
//  emitInterpOne(Type::Cell);
}

void HhbcTranslator::emitReqMod(const StringData* name) {
//  PUNT(ReqMod);
  emitReqDoc(name);
}

void HhbcTranslator::emitReqSrc(const StringData* name) {
//  PUNT(ReqSrc);
  emitReqDoc(name);
}

template<class Lambda>
SSATmp* HhbcTranslator::emitIterInitCommon(int offset, Lambda genFunc) {
  SSATmp* src = popC();
  Type type = src->getType();
  if (!type.isArray() && type != Type::Obj) {
    PUNT(IterInit);
  }
  spillStack();
  SSATmp* res = genFunc(src);
  return emitJmpCondHelper(offset, true, res);
}

void HhbcTranslator::emitIterInit(uint32_t iterId,
                                  int offset,
                                  uint32_t valLocalId) {
  emitIterInitCommon(offset, [=] (SSATmp* src) {
    return m_tb->genIterInit(src, iterId, valLocalId);
  });
}

void HhbcTranslator::emitIterInitK(uint32_t iterId,
                                   int offset,
                                   uint32_t valLocalId,
                                   uint32_t keyLocalId) {
  emitIterInitCommon(offset, [=] (SSATmp* src) {
    return m_tb->genIterInitK(src, iterId, valLocalId, keyLocalId);
  });
}

void HhbcTranslator::emitIterNext(uint32_t iterId,
                                  int offset,
                                  uint32_t valLocalId) {
  SSATmp* res = m_tb->genIterNext(iterId, valLocalId);
  emitJmpCondHelper(offset, false, res);
}

void HhbcTranslator::emitIterNextK(uint32_t iterId,
                                   int offset,
                                   uint32_t valLocalId,
                                   uint32_t keyLocalId) {
  SSATmp* res = m_tb->genIterNextK(iterId, valLocalId, keyLocalId);
  emitJmpCondHelper(offset, false, res);
}

void HhbcTranslator::emitIterFree(uint32_t iterId) {
  m_tb->genIterFree(iterId);
}

void HhbcTranslator::emitCreateCont(bool getArgs,
                                    Id funNameStrId) {
  emitInterpOneOrPunt(Type::Cell);
  return;
  // Task #2036200: Fix and re-enable this

  /* Runtime-determined slow path punts to TranslatorX64 for now */
  m_tb->genExitOnVarEnv(getExitSlowTrace());

  const StringData* genName = lookupStringId(funNameStrId);
  const Func* origFunc = getCurFunc();
  const Func* genFunc = origFunc->getGeneratorBody(genName);
  int origLocals = origFunc->numLocals();
  int genLocals = genFunc->numLocals();

  TCA helper = getCurFunc()->isNonClosureMethod() ?
    (TCA)&VMExecutionContext::createContinuation<true> :
    (TCA)&VMExecutionContext::createContinuation<false>;
  SSATmp* cont = m_tb->gen(CreateCont, cns(helper), m_tb->getFP(),
                           cns(getArgs), cns(origFunc), cns(genFunc));

  TranslatorX64::ContParamMap params;
  if (origLocals <= TranslatorX64::kMaxInlineContLocals &&
      TranslatorX64::mapContParams(params, origFunc, genFunc)) {
    static const StringData* thisStr = StringData::GetStaticString("this");
    Id thisId = kInvalidId;
    bool fillThis = origFunc->isNonClosureMethod() && !origFunc->isStatic() &&
      ((thisId = genFunc->lookupVarId(thisStr)) != kInvalidId) &&
      (origFunc->lookupVarId(thisStr) == kInvalidId);
    SSATmp* locals = m_tb->gen(LdContLocalsPtr, cont);
    for (int i = 0; i < origLocals; ++i) {
      SSATmp* loc = m_tb->genIncRef(m_tb->genLdLoc(i));
      m_tb->genStMem(locals, cellsToBytes(genLocals - params[i] - 1), loc,
                     true);
    }
    if (fillThis) {
      assert(thisId != kInvalidId);
      m_tb->gen(FillContThis, cont, locals,
                cns(cellsToBytes(genLocals - thisId - 1)));
    }
  } else {
    m_tb->gen(FillContLocals, m_tb->getFP(), cns(origFunc), cns(genFunc), cont);
  }

  push(cont);
}

void HhbcTranslator::emitContEnter(int32_t returnBcOffset) {
  spillStack();

  assert(getCurClass());
  SSATmp* cont = m_tb->genLdThis(nullptr);
  SSATmp* contAR = m_tb->genLdRaw(cont, RawMemSlot::ContARPtr, Type::StkPtr);

  SSATmp* func = m_tb->genLdARFuncPtr(contAR, m_tb->genDefConst<int64_t>(0));
  SSATmp* funcBody = m_tb->genLdRaw(func, RawMemSlot::FuncBody, Type::TCA);

  m_tb->genContEnter(contAR, funcBody, returnBcOffset);

  // We shouldn't need to change vmsp here
  assert(m_stackDeficit == 0);
}

void HhbcTranslator::emitContExit() {
  m_tb->genExitWhenSurprised(getExitSlowTrace());
  SSATmp* retAddr = m_tb->genLdRetAddr();
  // Despite the name, this doesn't actually free the AR; it updates the
  // hardware fp and returns the old one
  SSATmp* fp = m_tb->genFreeActRec();

  SSATmp* sp;
  if (m_stackDeficit) {
    // Adjust the hardware sp before leaving.
    // XXX This returns a ptrToGen
    sp = m_tb->genLdStackAddr(m_stackDeficit);
  } else {
    sp = m_tb->getSp();
  }

  m_tb->genRetCtrl(sp, fp, retAddr);
  m_hasExit = true;
}

void HhbcTranslator::emitUnpackCont() {
  m_tb->genLinkContVarEnv();
  SSATmp* cont = m_tb->genLdAssertedLoc(0, Type::Obj);
  push(m_tb->genLdRaw(cont, RawMemSlot::ContLabel, Type::Int));
}

void HhbcTranslator::emitPackCont(int64_t labelId) {
  m_tb->genUnlinkContVarEnv();
  SSATmp* cont = m_tb->genLdAssertedLoc(0, Type::Obj);
  m_tb->genSetPropCell(cont, CONTOFF(m_value), popC());
  m_tb->genStRaw(cont, RawMemSlot::ContLabel,
                 m_tb->genDefConst(labelId));
}

void HhbcTranslator::emitContReceive() {
  SSATmp* cont = m_tb->genLdAssertedLoc(0, Type::Obj);
  m_tb->genContRaiseCheck(cont, getExitSlowTrace());
  SSATmp* valOffset = m_tb->genDefConst<int64_t>(CONTOFF(m_received));
  SSATmp* value = m_tb->genLdProp(cont, valOffset, Type::Cell, nullptr);
  value = m_tb->genIncRef(value);
  push(value);
}

void HhbcTranslator::emitContRaised() {
  SSATmp* cont = m_tb->genLdAssertedLoc(0, Type::Obj);
  m_tb->genContRaiseCheck(cont, getExitSlowTrace());
}

void HhbcTranslator::emitContDone() {
  SSATmp* cont = m_tb->genLdAssertedLoc(0, Type::Obj);
  m_tb->genStRaw(cont, RawMemSlot::ContDone, m_tb->genDefConst(true));
  m_tb->genSetPropCell(cont, CONTOFF(m_value), m_tb->genDefInitNull());
}

void HhbcTranslator::emitContNext() {
  assert(getCurClass());
  SSATmp* cont = m_tb->genLdThis(nullptr);
  m_tb->genContPreNext(cont, getExitSlowTrace());
  m_tb->genSetPropCell(cont, CONTOFF(m_received), m_tb->genDefInitNull());
}

void HhbcTranslator::emitContSendImpl(bool raise) {
  assert(getCurClass());
  SSATmp* cont = m_tb->genLdThis(nullptr);
  m_tb->genContStartedCheck(cont, getExitSlowTrace());
  m_tb->genContPreNext(cont, getExitSlowTrace());

  SSATmp* value = m_tb->genLdAssertedLoc(0, Type::Cell);
  value = m_tb->genIncRef(value);
  m_tb->genSetPropCell(cont, CONTOFF(m_received), value);
  if (raise) {
    m_tb->genStRaw(cont, RawMemSlot::ContShouldThrow,
                   m_tb->genDefConst(true));
  }
}

void HhbcTranslator::emitContSend() {
  emitContSendImpl(false);
}

void HhbcTranslator::emitContRaise() {
  emitContSendImpl(true);
}

void HhbcTranslator::emitContValid() {
  assert(getCurClass());
  SSATmp* cont = m_tb->genLdThis(nullptr);
  SSATmp* done = m_tb->genLdRaw(cont, RawMemSlot::ContDone, Type::Bool);
  push(m_tb->genNot(done));
}

void HhbcTranslator::emitContCurrent() {
  assert(getCurClass());
  SSATmp* cont = m_tb->genLdThis(nullptr);
  m_tb->genContStartedCheck(cont, getExitSlowTrace());
  SSATmp* offset = m_tb->genDefConst<int64_t>(CONTOFF(m_value));
  SSATmp* value = m_tb->genLdProp(cont, offset, Type::Cell, nullptr);
  value = m_tb->genIncRef(value);
  push(value);
}

void HhbcTranslator::emitContStopped() {
  assert(getCurClass());
  SSATmp* cont = m_tb->genLdThis(nullptr);
  m_tb->genStRaw(cont, RawMemSlot::ContRunning, m_tb->genDefConst(false));
}

void HhbcTranslator::emitContHandle() {
  // No reason to punt, translator-x64 does emitInterpOne as well
  emitInterpOne(Type::None, 1);
}

void HhbcTranslator::emitStrlen() {
  Type inType = topC()->getType();

  if (inType.isString()) {
    SSATmp* input = popC();
    if (input->isConst()) {
      // static string; fold its strlen operation
      push(m_tb->genDefConst<int64_t>(input->getValStr()->size()));
    } else {
      push(m_tb->genLdRaw(input, RawMemSlot::StrLen, Type::Int));
      m_tb->genDecRef(input);
    }
  } else if (inType.isNull()) {
    popC();
    push(m_tb->genDefConst<int64_t>(0));
  } else if (inType == Type::Bool) {
    // strlen(true) == 1, strlen(false) == 0.
    push(m_tb->genConvToInt(popC()));
  } else {
    emitInterpOne(Type::Int | Type::InitNull, 1);
  }
}

void HhbcTranslator::emitIncStat(int32_t counter, int32_t value, bool force) {
  TRACE(3, "%u: IncStat %d %d\n", m_bcOff, counter, value);
  if (Stats::enabled() || force) {
    m_tb->genIncStat(counter, value, force);
  }
}

SSATmp* HhbcTranslator::getStrName(const StringData* knownName) {
  SSATmp* name = popC();
  assert(name->isA(Type::Str) || knownName);
  if (!name->isConst() || !name->isA(Type::Str)) {
    if (knownName) {
      // The SSATmp on the evaluation stack was not a string constant,
      // but the bytecode translator somehow knew the name statically.
      name = m_tb->genDefConst(knownName);
    }
  } else {
    assert(!knownName || knownName->same(name->getValStr()));
  }
  return name;
}

SSATmp* HhbcTranslator::emitLdClsPropAddrOrExit(const StringData* propName,
                                                Block* block) {
  SSATmp* clsTmp = popA();
  SSATmp* prop = getStrName(propName);
  SSATmp* addr = m_tb->gen(LdClsPropAddr,
                           block,
                           clsTmp,
                           prop,
                           m_tb->genDefConst(getCurClass()));
  m_tb->genDecRef(prop); // safe to do early because prop is a string
  return addr;
}

bool HhbcTranslator::checkSupportedClsProp(const StringData* propName,
                                           Type resultType,
                                           int stkIndex) {
  if (topC(stkIndex + 1)->isA(Type::Str) || propName) {
    return true;
  }
  emitInterpOne(resultType, stkIndex + 2);
  return false;
}

bool HhbcTranslator::checkSupportedGblName(const StringData* gblName,
                                           Type resultType,
                                           int stkIndex) {
  if (topC(stkIndex)->isA(Type::Str) || gblName) {
    return true;
  }
  emitInterpOne(resultType, stkIndex + 1);
  return false;
}

SSATmp* HhbcTranslator::emitLdGblAddr(const StringData* gblName, Block* block) {
  SSATmp* name = getStrName(gblName);
  // Note: Once we use control flow to implement IssetG/EmptyG, we can
  // use a LdGblAddr helper that decrefs name for us
  SSATmp* addr = m_tb->gen(LdGblAddr, block, name);
  m_tb->genDecRef(name);
  return addr;
}

SSATmp* HhbcTranslator::emitLdGblAddrDef(const StringData* gblName) {
  return m_tb->gen(LdGblAddrDef, getStrName(gblName));
}

void HhbcTranslator::emitIncDecS(bool pre, bool inc) {
  if (!checkSupportedClsProp(nullptr, Type::Cell, 0)) return;
  Trace* exit = getExitSlowTrace();
  emitIncDecMem(pre, inc, emitLdClsPropAddr(nullptr), exit);
}

void HhbcTranslator::emitIncDecProp(bool pre, bool inc, int offset,
                                    bool isPropOnStack) {
  TRACE(3, "%u: IncDecProp %d\n", m_bcOff, offset);
  Trace* exitTrace = getExitSlowTrace();
  if (isPropOnStack) {
    UNUSED SSATmp* prop = popC();
  }
  SSATmp* obj = popC();
  if (obj->getType() != Type::Obj) {
    PUNT(IncDecProp_nonobj);
  }
  SSATmp* propOffset = m_tb->genDefConst<int64_t>(offset);
  SSATmp* src = m_tb->genLdProp(obj, propOffset, Type::Int, exitTrace);
  // do the add and store back
  SSATmp* res = emitIncDec(pre, inc, src);
  m_tb->genStProp(obj, propOffset, res, false);
  m_tb->genDecRef(obj);
  m_tb->genDecRef(src);
}

void HhbcTranslator::emitSetProp(int offset, bool isPropOnStack) {
  // object + offset
  TRACE(3, "%u: SetProp %d\n", m_bcOff, offset);
  Trace* exitTrace = getExitSlowTrace();
  SSATmp* src = popC();
  if (isPropOnStack) {
    UNUSED SSATmp* prop = popC();
  }
  SSATmp* obj = popC();
  if (obj->getType() != Type::Obj) {
    PUNT(SetProp_nonobj);
  }
  SSATmp* propOffset = m_tb->genDefConst<int64_t>(offset);
  SSATmp* propPtr = m_tb->genUnboxPtr(m_tb->genLdPropAddr(obj, propOffset));
  SSATmp* prevValue = m_tb->genLdMem(propPtr, Type::Cell, exitTrace);
  m_tb->genStMem(propPtr, src, true);
  pushIncRef(src);
  m_tb->genDecRef(prevValue);
  m_tb->genDecRef(obj);
}

void HhbcTranslator::emitMInstr(const NormalizedInstruction& ni) {
  VectorTranslator(ni, *this).emit();
}

void HhbcTranslator::emitCGetProp(LocationCode locCode,
                                  int offset,
                                  bool isPropOnStack,
                                  Type resultType,
                                  bool isInferedType) {
  TRACE(3, "%u: CGetM %d\n", m_bcOff, offset);

  Trace* exitTrace1 = getExitSlowTrace();
  Trace* exitTrace2 = getExitSlowTrace();
  if (isPropOnStack) {
    UNUSED SSATmp* prop = popC();
  }
  SSATmp* obj;
  switch (locCode) {
    case LH:
      assert(getCurClass());
      obj = m_tb->genLdThis(nullptr);
      break;
    case LC:
      obj = popC();
      break;
    default:
      not_reached();
  }
  assert(obj->isA(Type::Obj));
  SSATmp* propOffset = m_tb->genDefConst<int64_t>(offset);
  SSATmp* val;

  if (isInferedType) {
    assert(resultType.isKnownUnboxedDataType());
    val = m_tb->genLdProp(obj, propOffset, resultType, nullptr);
  } else {
    if (resultType == Type::None) {
      // result type not predicted
      resultType = Type::Cell;
    } else {
      assert(resultType.isKnownUnboxedDataType());
    }
    // This code is currently correct, but once we enable type
    // prediction for CGetM, we should exit normally to a trace
    // executes the CGet as cell type (including the incref of the
    // result and decref of the obj) and exits to the next bytecode
    SSATmp* propPtr = m_tb->genUnboxPtr(m_tb->genLdPropAddr(obj, propOffset));
    val = m_tb->genLdMem(propPtr, resultType, exitTrace1);
    m_tb->gen(CheckInit, exitTrace2, val);
  }
  pushIncRef(val);
  if (locCode != LH) {
    m_tb->genDecRef(obj);
  }
}

/*
 * IssetH: return true if var is not uninit and !is_null(var)
 * Unboxes var if necessary when var is not uninit.
 */
void HhbcTranslator::emitIssetL(int32_t id) {
  TRACE(3, "%u: IssetL %d\n", m_bcOff, id);

  Type trackedType = m_tb->getLocalType(id);
  // guards should ensure we have type info at this point
  assert(trackedType != Type::None);
  if (trackedType.subtypeOf(Type::Uninit)) {
    push(m_tb->genDefConst(false));
  } else {
    Trace* exitTrace = getExitTrace();
    SSATmp* ld = m_tb->genLdLocAsCell(id, exitTrace);
    push(m_tb->gen(IsNType, Type::Null, ld));
  }
}

void HhbcTranslator::emitIssetG(const StringData* gblName) {
  TRACE(3, "%u: IssetG\n", m_bcOff);
  emitIsset(gblName,
            &HPHP::VM::JIT::HhbcTranslator::checkSupportedGblName,
            &HPHP::VM::JIT::HhbcTranslator::emitLdGblAddr);
}

void HhbcTranslator::emitIssetS(const StringData* propName) {
  TRACE(3, "%u: IssetS\n", m_bcOff);
  emitIsset(propName,
            &HPHP::VM::JIT::HhbcTranslator::checkSupportedClsProp,
            &HPHP::VM::JIT::HhbcTranslator::emitLdClsPropAddrOrExit);
}

void HhbcTranslator::emitEmptyL(int32_t id) {
  TRACE(3, "%u: EmptyL %d\n", m_bcOff, id);

  Type trackedType = m_tb->getLocalType(id);
  assert(trackedType != Type::None);
  if (trackedType == Type::Uninit) {
    push(m_tb->genDefConst(true));
  } else {
    Trace* exitTrace = getExitTrace();
    SSATmp* ld = m_tb->genLdLocAsCell(id, exitTrace);
    push(m_tb->genNot(m_tb->genConvToBool(ld)));
  }
}

void HhbcTranslator::emitEmptyG(const StringData* gblName) {
  TRACE(3, "%u: EmptyG\n", m_bcOff);
  emitEmpty(gblName,
            &HPHP::VM::JIT::HhbcTranslator::checkSupportedGblName,
            &HPHP::VM::JIT::HhbcTranslator::emitLdGblAddr);
}

void HhbcTranslator::emitEmptyS(const StringData* propName) {
  TRACE(3, "%u: EmptyS\n", m_bcOff);
  emitEmpty(propName,
            &HPHP::VM::JIT::HhbcTranslator::checkSupportedClsProp,
            &HPHP::VM::JIT::HhbcTranslator::emitLdClsPropAddrOrExit);
}

void HhbcTranslator::emitIsTypeC(Type t) {
  TRACE(3, "%u: Is%sC\n", m_bcOff, t.toString().c_str());
  SSATmp* src = popC();
  push(m_tb->gen(IsType, t, src));
  m_tb->genDecRef(src);
}

void HhbcTranslator::emitIsTypeL(Type t, int id) {
  TRACE(3, "%u: Is%sH\n", m_bcOff, t.toString().c_str());
  Trace* exitTrace = getExitTrace();
  push(m_tb->gen(IsType, t, emitLdLocWarn(id, exitTrace)));
}

void HhbcTranslator::emitIsNullL(int id)   { emitIsTypeL(Type::Null, id);}
void HhbcTranslator::emitIsArrayL(int id)  { emitIsTypeL(Type::Arr, id); }
void HhbcTranslator::emitIsStringL(int id) { emitIsTypeL(Type::Str, id); }
void HhbcTranslator::emitIsObjectL(int id) { emitIsTypeL(Type::Obj, id); }
void HhbcTranslator::emitIsIntL(int id)    { emitIsTypeL(Type::Int, id); }
void HhbcTranslator::emitIsBoolL(int id)   { emitIsTypeL(Type::Bool, id);}
void HhbcTranslator::emitIsDoubleL(int id) { emitIsTypeL(Type::Dbl, id); }
void HhbcTranslator::emitIsNullC()   { emitIsTypeC(Type::Null);}
void HhbcTranslator::emitIsArrayC()  { emitIsTypeC(Type::Arr); }
void HhbcTranslator::emitIsStringC() { emitIsTypeC(Type::Str); }
void HhbcTranslator::emitIsObjectC() { emitIsTypeC(Type::Obj); }
void HhbcTranslator::emitIsIntC()    { emitIsTypeC(Type::Int); }
void HhbcTranslator::emitIsBoolC()   { emitIsTypeC(Type::Bool);}
void HhbcTranslator::emitIsDoubleC() { emitIsTypeC(Type::Dbl); }

void HhbcTranslator::emitPopC() {
  TRACE(3, "%u: PopC\n", m_bcOff);
  popDecRef(Type::Cell);
}

void HhbcTranslator::emitPopV() {
  TRACE(3, "%u: PopV\n", m_bcOff);
  popDecRef(Type::BoxedCell);
}

void HhbcTranslator::emitPopR() {
  TRACE(3, "%u: PopR\n", m_bcOff);
  popDecRef(Type::Gen);
}

void HhbcTranslator::emitDup() {
  TRACE(3, "%u: Dup\n", m_bcOff);
  pushIncRef(topC());
}

void HhbcTranslator::emitJmp(int32_t offset,
                             bool  breakTracelet,
                             bool  noSurprise) {
  TRACE(3, "%u: Jmp %d\n", m_bcOff, offset);
  // If surprise flags are set, exit trace and handle surprise
  bool backward = (offset - (int32_t)m_bcOff) < 0;
  if (backward && !noSurprise) {
    m_tb->genExitWhenSurprised(getExitSlowTrace());
  }
  if (!breakTracelet) return;
  m_tb->genJmp(getExitTrace(offset));
}

SSATmp* HhbcTranslator::emitJmpCondHelper(int32_t offset,
                                         bool negate,
                                         SSATmp* src) {
  Trace* target = nullptr;
  if (m_lastBcOff) {
    // Spill everything on main trace if all paths will exit
    spillStack();
    target = getExitTrace(offset, getBcOffNextTrace());
  } else {
    target = getExitTrace(offset);
  }
  SSATmp* boolSrc = m_tb->genConvToBool(src);
  m_tb->genDecRef(src);
  return m_tb->genJmpCond(boolSrc, target, negate);
}

void HhbcTranslator::emitJmpZ(int32_t offset) {
  TRACE(3, "%u: JmpZ %d\n", m_bcOff, offset);
  SSATmp* src = popC();
  emitJmpCondHelper(offset, true, src);
}

void HhbcTranslator::emitJmpNZ(int32_t offset) {
  TRACE(3, "%u: JmpNZ %d\n", m_bcOff, offset);
  SSATmp* src = popC();
  emitJmpCondHelper(offset, false, src);
}

void HhbcTranslator::emitCmp(Opcode opc) {
  TRACE(3, "%u: Cmp %s\n", m_bcOff, opcodeName(opc));

  if (cmpOpTypesMayReenter(opc, topC(0)->getType(), topC(1)->getType())) {
    spillStack();
  }
  // src2 opc src1
  SSATmp* src1 = popC();
  SSATmp* src2 = popC();
  push(m_tb->genCmp(opc, src2, src1));
  m_tb->genDecRef(src2);
  m_tb->genDecRef(src1);
}

void HhbcTranslator::emitClsCnsD(int32_t cnsNameStrId, int32_t clsNameStrId) {
  // This bytecode re-enters if there is no class with the given name
  // and can throw a fatal error.
  const StringData* cnsNameStr = lookupStringId(cnsNameStrId);
  const StringData* clsNameStr = lookupStringId(clsNameStrId);
  TRACE(3, "%u: ClsCnsD %s::%s\n", m_bcOff, clsNameStr->data(),
        cnsNameStr->data());
  SSATmp* cnsNameTmp = m_tb->genDefConst(cnsNameStr);
  SSATmp* clsNameTmp = m_tb->genDefConst(clsNameStr);
  if (0) {
    // TODO: 2068502 pick one of these two implementations and remove the other.
    Trace* exitTrace = getExitSlowTrace();
    SSATmp* cns = m_tb->gen(LdClsCns, Type::Cell, cnsNameTmp, clsNameTmp);
    m_tb->genCheckInit(cns, exitTrace);
    push(cns);
  } else {
    // if-then-else
    // todo: t2068502: refine the type? hhbc spec says null|bool|int|dbl|str
    //       and, str should always be static-str.
    spillStack(); // do this on main trace so we update stack tracking once.
    Type cnsType = Type::Cell;
    SSATmp* c1 = m_tb->gen(LdClsCns, cnsType, cnsNameTmp, clsNameTmp);
    SSATmp* result = m_tb->cond(getCurFunc(),
      [&] (Block* taken) { // branch
        m_tb->genCheckInit(c1, taken);
      },
      [&] { // Next: LdClsCns hit in TC
        return c1;
      },
      [&] { // Taken: miss in TC, do lookup & init
        m_tb->hint(Block::Unlikely);
        return m_tb->gen(LookupClsCns, cnsType, cnsNameTmp, clsNameTmp);
      }
    );
    push(result);
  }
}

void HhbcTranslator::emitAKExists() {
  TRACE(3, "%u: AKExists\n", m_bcOff);
  SSATmp* arr = popC();
  SSATmp* key = popC();

  if (!arr->isA(Type::Arr) && !arr->isA(Type::Obj)) {
    PUNT(AKExists_badArray);
  }
  if (!key->isString() && !key->isA(Type::Int) && !key->isA(Type::Null)) {
    PUNT(AKExists_badKey);
  }

  push(m_tb->gen(AKExists, arr, key));
  m_tb->genDecRef(arr);
  m_tb->genDecRef(key);
}

void HhbcTranslator::emitFPassR() {
  TRACE(3, "%u: FPassR\n", m_bcOff);
  emitUnboxRAux();
}

void HhbcTranslator::emitFPassCOp() {
  TRACE(3, "%u: FPassCOp\n", m_bcOff);
}

void HhbcTranslator::emitFPassV() {
  TRACE(3, "%u: FPassV\n", m_bcOff);
  Block* exit = getExitTrace()->front();
  SSATmp* tmp = popV();
  pushIncRef(m_tb->gen(Unbox, exit, tmp));
  m_tb->genDecRef(tmp);
}

void HhbcTranslator::emitNativeImpl() {
  TRACE(3, "%u: NativeImpl\n", m_bcOff);
  m_tb->genNativeImpl();
  SSATmp* sp = m_tb->genRetAdjustStack();
  SSATmp* retAddr = m_tb->genLdRetAddr();
  SSATmp* fp = m_tb->genFreeActRec();    // updates fp
  m_tb->genRetCtrl(sp, fp, retAddr);

  // Flag that this trace has a Ret instruction so no ExitTrace is needed
  m_hasExit = true;
}

void HhbcTranslator::emitFPushCtor(int32_t numParams) {
  TRACE(3, "%u: FPushFuncCtor %d\n", m_bcOff, numParams);
  SSATmp* cls = popA();
  spillStack();
  SSATmp* newObj = m_tb->genNewObj(numParams, cls);
  m_fpiStack.push(newObj);
}

void HhbcTranslator::emitFPushCtorD(int32_t numParams, int32_t classNameStrId) {
  const StringData* className = lookupStringId(classNameStrId);
  TRACE(3, "%u: FPushFuncCtorD %d %s\n", m_bcOff, numParams, className->data());
  spillStack();
  SSATmp* newObj = m_tb->genNewObj(numParams, className);
  m_fpiStack.push(newObj);
}

void HhbcTranslator::emitFPushFuncD(int32_t numParams, int32_t funcId) {
  TRACE(3, "%u: FPushFuncD %d %d\n", m_bcOff, numParams, funcId);
  const NamedEntityPair& nep = lookupNamedEntityPairId(funcId);
  const StringData* name = nep.first;
  const Func* func       = Unit::lookupFunc(nep.second, name);
  // Translation is only supported if function lookup succeeds
  if (!func) {
    PUNT(FPushFuncDNull);
  }
  func->validate();

  const bool immutable = func->isNameBindingImmutable(getCurUnit());

  if (!immutable) {
    spillStack();  // LdFixedFunc can reenter
  }
  SSATmp* ssaFunc = immutable ? m_tb->genDefConst(func)
                              : m_tb->gen(LdFixedFunc,
                                          m_tb->genDefConst(name));
  SSATmp* actRec  = m_tb->genDefActRec(ssaFunc,
                                       m_tb->genDefInitNull(),
                                       numParams,
                                       nullptr);
  m_evalStack.push(actRec);
  spillStack(); // TODO(#2036900)
  m_fpiStack.push(immutable ? actRec : ssaFunc);
}

void HhbcTranslator::emitFPushFunc(int32_t numParams) {
  TRACE(3, "%u: FPushFuncD %d\n", m_bcOff, numParams);
  // input must be a string or an object implementing __invoke();
  // otherwise fatal
  SSATmp* funcName = popC();
  if (!funcName->isString()) {
    PUNT(FPushFunc_not_Str);
  }

  spillStack(); // LdFunc can reenter
  SSATmp* func = m_tb->gen(LdFunc, funcName);
  m_fpiStack.push(func);
  m_evalStack.push(m_tb->genDefActRec(func,
                                      m_tb->genDefInitNull(),
                                      numParams,
                                      nullptr));
  spillStack(); // TODO(#2036900)
}

void HhbcTranslator::emitFPushObjMethodD(int32_t numParams,
                                         int32_t methodNameStrId,
                                         const Class* baseClass) {
  const StringData* methodName = lookupStringId(methodNameStrId);
  TRACE(3, "%u: FPushObjMethodD %s %d\n",
        m_bcOff,
        methodName->data(),
        numParams);
  bool magicCall = false;
  SSATmp* funcTmp = nullptr;
  const Func* func = HPHP::VM::Transl::lookupImmutableMethod(baseClass,
                                                             methodName,
                                                             magicCall,
                                                         /* staticLookup: */
                                                             false);
  SSATmp* objOrCls = popC();

  if (!func) {
    if (baseClass && !(baseClass->attrs() & AttrInterface)) {
      MethodLookup::LookupResult res =
        g_vmContext->lookupObjMethod(func, baseClass, methodName, false);
      if ((res == MethodLookup::MethodFoundWithThis ||
           res == MethodLookup::MethodFoundNoThis) &&
          !func->isAbstract()) {
        /*
         * If we found the func in baseClass, then either:
         *  a) its private, and this is always going to be the
         *     called function. This case is handled further down.
         * OR
         *  b) any derived class must have a func that matches in staticness
         *     and is at least as accessible (and in particular, you can't
         *     override a public/protected method with a private method).
         *     In this case, we emit code to dynamically lookup the method
         *     given the Object and the method slot, which is the same as func's.
         */
        if (!(func->attrs() & AttrPrivate)) {
          SSATmp* clsTmp = m_tb->gen(LdObjClass, objOrCls);
          funcTmp = m_tb->genLdClsMethod(clsTmp, func->methodSlot());
          if (res == MethodLookup::MethodFoundNoThis) {
            m_tb->genDecRef(objOrCls);
            objOrCls = clsTmp;
          }
        }
      } else {
        func = nullptr; // force lookup
      }
    }
  }

  if (func != nullptr && funcTmp == nullptr) {
    if (func->attrs() & AttrStatic) {
      assert(baseClass);  // This assert may be too strong, but be aggressive
      // static function: store base class into this slot instead of obj
      // and decref the obj that was pushed as the this pointer since
      // the obj won't be in the actrec and thus MethodCache::lookup won't
      // decref it
      m_tb->genDecRef(objOrCls);
      objOrCls = m_tb->genDefConst(baseClass);
    }
  }

  const StringData* invName = nullptr;
  if (!funcTmp) {
    funcTmp = func ? m_tb->genDefConst(func)
                   : m_tb->genDefNull();
    if (func && magicCall) {
      invName = methodName;
    }
  }
  SSATmp* actRec = m_tb->genDefActRec(funcTmp,
                                      objOrCls,
                                      numParams,
                                      invName);
  m_evalStack.push(actRec);
  spillStack(); // TODO(#2036900)
  if (!func) {
    SSATmp* sp = spillStack();
    SSATmp* meth = m_tb->genLdObjMethod(methodName, sp);
    m_fpiStack.push(meth);
  } else {
    m_fpiStack.push(actRec);
  }
}

void HhbcTranslator::emitFPushClsMethodD(int32_t numParams,
                                         int32_t methodNameStrId,
                                         int32_t clssNamedEntityPairId) {

  const StringData* methodName = lookupStringId(methodNameStrId);
  const NamedEntityPair& np = lookupNamedEntityPairId(clssNamedEntityPairId);
  UNUSED const StringData* className = np.first;
  TRACE(3, "%u: FPushClsMethodD %s::%s %d\n", m_bcOff, className->data(),
        methodName->data(), numParams);
  const Class* baseClass = Unit::lookupClass(np.second);
  bool magicCall = false;
  const Func* func = HPHP::VM::Transl::lookupImmutableMethod(baseClass,
                                                             methodName,
                                                             magicCall,
                                                         /* staticLookup: */
                                                             true);
  bool mightNotBeStatic = false;
  if (func &&
      !(func->attrs() & AttrStatic) &&
      !(getCurFunc()->attrs() & AttrStatic) &&
      getCurClass() &&
      getCurClass()->classof(baseClass)) {
    mightNotBeStatic = true;
  }

  SSATmp* actRec;
  if (func) {
    SSATmp* objOrCls;
    if (!mightNotBeStatic) { // definitely static
      // static function: store base class into the m_cls/m_this slot
      objOrCls = m_tb->genDefConst(baseClass);
    } else if (m_tb->isThisAvailable()) {
      // 'this' pointer is available, so use it.
      assert(getCurClass());
      objOrCls = m_tb->genIncRef(m_tb->genLdThis(nullptr));
    } else {
      // might be a non-static call
      // generate code that tests at runtime whether to use
      // this pointer or class
      PUNT(FPushClsMethodD_MightNotBeStatic);
      assert(0);
    }
    actRec = m_tb->genDefActRec(m_tb->genDefConst<const Func*>(func),
                                objOrCls,
                                numParams,
                                func && magicCall ? methodName : nullptr);
  } else {
    // lookup static method & class in the target cache
    Trace* exitTrace = getExitSlowTrace();
    const StringData* className = np.first;
    SSATmp* funcClassTmp =
      m_tb->genLdClsMethodCache(m_tb->genDefConst(className),
                                m_tb->genDefConst(methodName),
                                m_tb->genDefConst(np.second),
                                exitTrace);
    actRec = m_tb->genDefActRec(funcClassTmp,
                                m_tb->genDefInitNull(),
                                numParams,
                                nullptr);
  }
  m_evalStack.push(actRec);
  spillStack(); // TODO(#2036900)
  m_fpiStack.push(actRec);
}

void HhbcTranslator::emitFPushClsMethodF(int32_t             numParams,
                                         const Class*      cls,
                                         const StringData* methName) {


  Block* exitBlock = getExitSlowTrace()->front();

  UNUSED SSATmp* clsVal  = popC();
  UNUSED SSATmp* methVal = popC();

  assert(cls);
  assert(methName && methName->isStatic());

  bool magicCall = false;
  const Func* func = lookupImmutableMethod(cls, methName, magicCall,
                                           true /* staticLookup */);
  UNUSED ActRec* fp = curFrame();
  assert(!fp->hasThis() || fp->getThis()->instanceof(cls));

  SSATmp* actRec = nullptr;
  SSATmp* curCtxTmp = m_tb->genLdCtx(getCurFunc());
  if (func) {
    SSATmp*   funcTmp = m_tb->genDefConst(func);
    SSATmp* newCtxTmp = m_tb->gen(GetCtxFwdCall, curCtxTmp, funcTmp);

    actRec = m_tb->genDefActRec(funcTmp, newCtxTmp, numParams,
                                (magicCall ? methName : nullptr));

  } else {
    SSATmp*      clsTmp = m_tb->genDefConst(cls);
    SSATmp* methNameTmp = m_tb->genDefConst(methName);
    SSATmp*  funcCtxTmp = m_tb->gen(LdClsMethodFCache, exitBlock, clsTmp,
                                    methNameTmp, curCtxTmp);
    actRec = m_tb->genDefActRec(funcCtxTmp,
                                m_tb->genDefInitNull(),
                                numParams,
                                (magicCall ? methName : nullptr));
  }
  m_evalStack.push(actRec);
  spillStack(); // TODO(#2036900)
  m_fpiStack.push(actRec);
}

void HhbcTranslator::emitFCallArray() {
  PUNT(FCallArray); // can't interpret one because of control flow
}

void HhbcTranslator::emitFCall(uint32_t numParams,
                               Offset returnBcOffset,
                               const Func* callee) {
  // pop the actrec or func from FPI stack
  SSATmp* actRecOrFunc = m_fpiStack.pop();

  // pop the incoming parameters to the call
  SSATmp* params[numParams];
  for (uint32_t i = 0; i < numParams; i++) {
    params[numParams - i - 1] = popF();
  }

  SSATmp* func = callee ? m_tb->genDefConst(callee) : nullptr;
  SSATmp* actRec = spillStack();
  if (!func) {
    if (actRecOrFunc && actRecOrFunc->getType() == Type::Func) {
      func = actRecOrFunc;
    } else {
      func = m_tb->genDefNull();
    }
  }

  m_tb->genCall(actRec,
                returnBcOffset,
                func,
                numParams,
                params);
}

// This is used to check that return types of builtins are not simple
// types. This is different from IS_REFCOUNTED_TYPE because builtins
// can return Variants, and we use KindOfUnknown to denote these
// return types.
static bool isCppByRef(DataType t) {
  assert(t != KindOfDouble);
  return t != KindOfBoolean && t != KindOfInt64 && t != KindOfNull;
}

void HhbcTranslator::emitFCallBuiltin(uint32_t numArgs,
                                      uint32_t numNonDefault, int32_t funcId) {
  const NamedEntityPair& nep = lookupNamedEntityPairId(funcId);
  const StringData* name = nep.first;
  const Func* callee = Unit::lookupFunc(nep.second, name);

  callee->validate();

  // spill args to stack. We need to spill these for two resons:
  // 1. some of the arguments may be passed by reference, for which
  //    case we will generate LdStackAddr() (see below).
  // 2. type conversions of the arguments (using tvCast* helpers)
  //    may throw an exception, so we need to have the VM stack
  //    in a clean state at that point.
  spillStack();
  // Convert types if needed
  for (int i = 0; i < numNonDefault; i++) {
    const Func::ParamInfo& pi = callee->params()[i];
    switch (pi.builtinType()) {
      case KindOfBoolean:
      case KindOfInt64:
      case KindOfArray:
      case KindOfObject:
      case KindOfString:
        m_tb->genCastStk(numArgs - i - 1,
                         Type::fromDataType(pi.builtinType(), KindOfInvalid));
        break;
      case KindOfDouble: not_reached();
      case KindOfUnknown: break;
      default:            not_reached();
    }
  }


  // pass arguments for call
  SSATmp* args[numArgs + 1];

  int isRefReturn = 0;
  // if the function returns by reference, the first parameter
  // has to be the fixed C++ location for the return value
  if (isCppByRef(callee->returnType())) {
    isRefReturn = 1;
    SSATmp* misBase = m_tb->gen(DefMIStateBase);
    SSATmp* returnAddr = m_tb->genLdAddr(misBase, HHIR_MISOFF(tvBuiltinReturn));
    args[0] = returnAddr;
  }

  for (int i = numArgs - 1; i >= 0; i--) {
    const Func::ParamInfo& pi = callee->params()[i];
    switch (pi.builtinType()) {
      case KindOfBoolean:
      case KindOfInt64:
        args[isRefReturn + i] = top(Type::fromDataType(pi.builtinType(), KindOfInvalid),
                                    numArgs - i - 1);
        break;
      case KindOfDouble: assert(false);
      default:
        args[isRefReturn + i] = loadStackAddr(numArgs - i - 1);
        break;
    }
  }
  // generate call and set return type
  SSATmp* func = m_tb->genDefConst<const Func*>(callee);
  Type type = Type::fromDataTypeWithRef(callee->returnType(),
                       (callee->attrs() & ClassInfo::IsReference));
  SSATmp* ret = m_tb->genCallBuiltin(func, type, numArgs + isRefReturn, args);

  // decref and free args
  for (int i = 0; i < numArgs; i++) {
    SSATmp* arg = popR();
    if (i >= numArgs - numNonDefault) {
      m_tb->genDecRef(arg);
    }
  }

  // push return value
  push(ret);
}

void HhbcTranslator::emitRet(Type type, bool freeInline) {
  const Func* curFunc = getCurFunc();
  bool mayUseVV = (curFunc->attrs() & AttrMayUseVV);
  bool mayHaveThis = (curFunc->isPseudoMain() ||
                      (curFunc->isMethod() && !curFunc->isStatic()));

  m_tb->genExitWhenSurprised(getExitSlowTrace());
  if (mayUseVV) {
    // Note: this has to be the first thing, because we cannot bail after
    //       we start decRefing locs because then there'll be no corresponding
    //       bytecode boundaries until the end of RetC
    m_tb->genReleaseVVOrExit(getExitSlowTrace());
  }
  SSATmp* retVal = pop(type);
  if (mayHaveThis) {
    m_tb->genDecRefThis();
  }

  SSATmp* sp;
  if (freeInline) {
    for (int id = curFunc->numLocals() - 1; id >= 0; --id) {
      /*
       * TODO(#1980291): this doesn't correctly handle
       * debug_backtrace.
       */
      m_tb->genDecRefLoc(id);
    }
    m_tb->genRetVal(retVal);
    sp = m_tb->genRetAdjustStack();
  } else {
    sp = m_tb->genGenericRetDecRefs(retVal, curFunc->numLocals());
    m_tb->genRetVal(retVal);
  }

  // Free ActRec, and return control to caller.
  SSATmp* retAddr = m_tb->genLdRetAddr();
  SSATmp* fp = m_tb->genFreeActRec();
  m_tb->genRetCtrl(sp, fp, retAddr);

  // Flag that this trace has a Ret instruction, so that no ExitTrace is needed
  m_hasExit = true;
}

void HhbcTranslator::emitSwitch(const ImmVector& iv,
                                int64_t base,
                                bool bounded) {
  int nTargets = bounded ? iv.size() - 2 : iv.size();

  SSATmp* const switchVal = popC();
  Type type = switchVal->getType();
  assert(IMPLIES(!type.equals(Type::Int), bounded));
  assert(IMPLIES(bounded, iv.size() > 2));
  SSATmp* index;
  SSATmp* ssabase = m_tb->genDefConst(base);
  SSATmp* ssatargets = m_tb->genDefConst(nTargets);

  Offset defaultOff = m_bcOff + iv.vec32()[iv.size() - 1];
  Offset zeroOff = 0;
  if (base <= 0 && (base + nTargets) > 0) {
    zeroOff = m_bcOff + iv.vec32()[0 - base];
  } else {
    zeroOff = defaultOff;
  }

  if (type.subtypeOf(Type::Null)) {
    m_tb->genJmp(getExitTrace(zeroOff));
    return;
  } else if (type.subtypeOf(Type::Bool)) {
    Offset nonZeroOff = m_bcOff + iv.vec32()[iv.size() - 2];
    m_tb->genJmpCond(switchVal, getExitTrace(nonZeroOff), false);
    m_tb->genJmp(getExitTrace(zeroOff));
    return;
  } else if (type.subtypeOf(Type::Int)) {
    // No special treatment needed
    index = switchVal;
  } else if (type.subtypeOf(Type::Dbl)) {
    // switch(Double|String|Obj)Helper do bounds-checking for us, so
    // we need to make sure the default case is in the jump table,
    // and don't emit our own bounds-checking code
    bounded = false;
    index = m_tb->gen(LdSwitchDblIndex,
                      switchVal, ssabase, ssatargets);
  } else if (type.subtypeOf(Type::Str)) {
    bounded = false;
    index = m_tb->gen(LdSwitchStrIndex,
                      switchVal, ssabase, ssatargets);
  } else if (type.subtypeOf(Type::Obj)) {
    // switchObjHelper can throw exceptions and reenter the VM
    if (type.subtypeOf(Type::Obj)) {
      spillStack();
    }
    bounded = false;
    index = m_tb->gen(LdSwitchObjIndex,
                      switchVal, ssabase, ssatargets);
  } else if (type.subtypeOf(Type::Arr)) {
    m_tb->genDecRef(switchVal);
    m_tb->genJmp(getExitTrace(defaultOff));
    return;
  } else {
    PUNT(Switch-UnknownType);
  }

  std::vector<Offset> targets(iv.size());
  for (int i = 0; i < iv.size(); i++) {
    targets[i] = m_bcOff + iv.vec32()[i];
  }

  JmpSwitchData data;
  data.func        = getCurFunc();
  data.base        = base;
  data.bounded     = bounded;
  data.cases       = iv.size();
  data.defaultOff  = defaultOff;
  data.targets     = &targets[0];

  SSATmp* stack = spillStack();
  m_tb->gen(SyncVMRegs, m_tb->getFp(), stack);

  m_tb->gen(JmpSwitchDest, &data, index);
  m_hasExit = true;
}

void HhbcTranslator::emitSSwitch(const ImmVector& iv) {
  const int numCases = iv.size() - 1;

  /*
   * We use a fast path translation with a hashtable if none of the
   * cases are numeric strings and if the input is actually a string.
   *
   * Otherwise we do a linear search through the cases calling string
   * conversion routines.
   */
  const bool fastPath =
    topC()->isA(Type::Str) &&
    std::none_of(iv.strvec(), iv.strvec() + numCases,
      [&](const StrVecItem& item) {
        return getCurUnit()->lookupLitstrId(item.str)->isNumeric();
      }
    );

  // The slow path can throw exceptions and reenter the VM.
  if (!fastPath) spillStack();

  SSATmp* const testVal = popC();
  assert(m_bcOff != -1);
  auto const bcOff = m_bcOff;

  std::vector<LdSSwitchData::Elm> cases(numCases);
  for (int i = 0; i < numCases; ++i) {
    auto const& kv = iv.strvec()[i];
    cases[i].str  = getCurUnit()->lookupLitstrId(kv.str);
    cases[i].dest = bcOff + kv.dest;
  }

  LdSSwitchData data;
  data.func       = getCurFunc();
  data.numCases   = numCases;
  data.cases      = &cases[0];
  data.defaultOff = bcOff + iv.strvec()[iv.size() - 1].dest;

  SSATmp* dest = m_tb->gen(fastPath ? LdSSwitchDestFast
                                    : LdSSwitchDestSlow,
                           &data,
                           testVal);
  m_tb->genDecRef(testVal);
  SSATmp* stack = spillStack();
  m_tb->gen(SyncVMRegs, m_tb->getFp(), stack);
  m_tb->gen(JmpIndirect, dest);
  m_hasExit = true;
}

void HhbcTranslator::emitRetC(bool freeInline) {
  emitRet(Type::Cell, freeInline);
}

void HhbcTranslator::emitRetV(bool freeInline) {
  emitRet(Type::BoxedCell, freeInline);
}

void HhbcTranslator::setThisAvailable() {
  m_tb->setThisAvailable();
}

void HhbcTranslator::guardTypeLocal(uint32_t locId, Type type) {
  checkTypeLocal(locId, type);
  m_typeGuards.push_back(TypeGuard(TypeGuard::Local, locId, type));
}

void HhbcTranslator::checkTypeLocal(uint32_t locId, Type type) {
  m_tb->genGuardLoc(locId, type, getExitTrace());
}

void HhbcTranslator::assertTypeLocal(uint32_t localIndex, Type type) {
  m_tb->genAssertLoc(localIndex, type);
}

Trace* HhbcTranslator::guardTypeStack(uint32_t stackIndex,
                                      Type type,
                                      Trace* nextTrace) {
  if (type.subtypeOf(Type::Cls)) {
    // Should not generate guards for class; instead assert their type
    assertTypeStack(stackIndex, type);
    return nextTrace;
  }
  if (nextTrace == nullptr) {
    nextTrace = getGuardExit();
  }
  m_tb->genGuardStk(stackIndex, type, nextTrace);
  m_typeGuards.push_back(TypeGuard(TypeGuard::Stack, stackIndex, type));

  return nextTrace;
}

void HhbcTranslator::checkTypeTopOfStack(Type type,
                                         Offset nextByteCode) {
  Trace* exitTrace = getExitTrace(nextByteCode);
  SSATmp* tmp = m_evalStack.top();
  if (!tmp) {
    FTRACE(1, "checkTypeTopOfStack: no tmp: {}\n", type.toString());
    m_tb->genGuardStk(0, type, exitTrace);
    push(pop(type));
  } else {
    FTRACE(1, "checkTypeTopOfStack: generating GuardType for {}\n",
           type.toString());
    m_evalStack.pop();
    tmp = m_tb->genGuardType(tmp, type, exitTrace);
    push(tmp);
  }
}

void HhbcTranslator::assertTypeStack(uint32_t stackIndex, Type type) {
  SSATmp* tmp = m_evalStack.top(stackIndex);
  if (!tmp) {
    m_tb->genAssertStk(stackIndex, type);
    return;
  }

  /*
   * We already had a value in flight---refine the type in case it
   * allows generating better code.  This is safe because in this path
   * we know the value is *actually* this type due to static analysis
   * (not based on guards).
   */
  refineType(tmp, type);
}

void HhbcTranslator::emitLoadDeps() {
  for (auto& guard : m_typeGuards) {
    switch (guard.getKind()) {
    case TypeGuard::Local:
      m_tb->genLdLoc(guard.getIndex());
      break;
    case TypeGuard::Stack:
      break;
    default:
      assert(false); // iterator guards should not happen
    }
  }
}

Trace* HhbcTranslator::guardRefs(int64_t               entryArDelta,
                                 const vector<bool>& mask,
                                 const vector<bool>& vals,
                                 Trace*              exitTrace) {
  if (exitTrace == nullptr) {
    exitTrace = getGuardExit();
  }

  int32_t actRecOff = cellsToBytes(entryArDelta);
  SSATmp* funcPtr = m_tb->genLdARFuncPtr(m_tb->getSp(),
                                         m_tb->genDefConst<int64_t>(actRecOff));
  SSATmp* nParams = m_tb->genLdRaw(funcPtr, RawMemSlot::FuncNumParams,
                                  Type::Int);
  SSATmp* bitsPtr = m_tb->genLdRaw(funcPtr, RawMemSlot::FuncRefBitVec,
                                  Type::Int);

  for (unsigned i = 0; i < mask.size(); i += 64) {
    assert(i < vals.size());

    uint64_t mask64 = TranslatorX64::packBitVec(mask, i);
    if (mask64 == 0) {
      continue;
    }

    uint64_t vals64 = TranslatorX64::packBitVec(vals, i);

    SSATmp* mask64Tmp   = m_tb->genLdConst<int64_t>(mask64);
    SSATmp* vals64Tmp   = m_tb->genLdConst<int64_t>(vals64);
    SSATmp* firstBitNum = m_tb->genDefConst<int64_t>(i);

    m_tb->genGuardRefs(funcPtr, nParams, bitsPtr, firstBitNum,
                      mask64Tmp, vals64Tmp, exitTrace);
  }

  return exitTrace;
}

void HhbcTranslator::emitVerifyParamType(int32_t paramId,
                                         const StringData* constraintClsName) {
  /* It's currently better to interpOne all of these, since the slow path punts
   * for the entire tracelet */
  emitInterpOneOrPunt(Type::None);
  return;

  assert(!constraintClsName || constraintClsName->isStatic());
  const Func* func = getCurFunc();
  const Class* constraint = (constraintClsName ?
                             Unit::lookupClass(constraintClsName) : nullptr);
  const TypeConstraint& tc = func->params()[paramId].typeConstraint();
  const StringData* tcTypeName = tc.typeName();

  // VerifyParamType does not remove the parameter from the stack
  TRACE(3, "%u: VerifyParamType %s\n", m_bcOff, tcTypeName->data());
  Trace* exitTrace2 = getExitSlowTrace();
  // XXX Should verify param type unbox?
  SSATmp* param = m_tb->genLdAssertedLoc(paramId, Type::Gen);
  if (param->getType() != Type::Obj) {
    PUNT(VerifyParamType_nonobj);
  }
  SSATmp* objClass = m_tb->gen(LdObjClass, param);

  if (tc.isObject()) {
    if (!(param->getType() == Type::Obj ||
          (param->getType().isNull() && tc.nullable()))) {
      emitInterpOne(Type::None);
      return;
    }
  } else {
    if (!tc.check(frame_local(curFrame() /* FIXME */, paramId), getCurFunc())) {
      emitInterpOne(Type::None);
      return;
    }
  }

  m_tb->genVerifyParamType(objClass,
                          m_tb->genDefConst(tcTypeName),
                          constraint,
                          exitTrace2);
}

void HhbcTranslator::emitInstanceOfD(int classNameStrId) {
  const StringData* className = lookupStringId(classNameStrId);
  SSATmp* src = popC();

  /*
   * InstanceOfD is always false if it's not an object.
   *
   * We're prepared to generate translations for known non-object
   * types, but if it's Gen/Cell we're going to PUNT because it's
   * natural to translate that case with control flow TODO(#2020251)
   */
  if (Type::Obj.strictSubtypeOf(src->getType())) {
    PUNT(InstanceOfD_MaybeObj);
  }
  if (!src->isA(Type::Obj)) {
    push(m_tb->genDefConst(false));
    m_tb->genDecRef(src);
    return;
  }

  SSATmp* objClass     = m_tb->gen(LdObjClass, src);
  SSATmp* ssaClassName = m_tb->genDefConst(className);

  Class::initInstanceBits();
  const bool haveBit = Class::haveInstanceBit(className);

  Class* const maybeCls = Unit::lookupClass(className);
  const bool isNormalClass = maybeCls &&
                             !(maybeCls->attrs() &
                               (AttrTrait | AttrInterface));
  const bool isUnique = RuntimeOption::RepoAuthoritative &&
                        maybeCls && (maybeCls->attrs() & AttrUnique);

  /*
   * If the class is unique or a parent of the current context, we
   * don't need to load it out of target cache because it must
   * already exist and be defined.
   *
   * Otherwise, we only use LdCachedClass---instanceof with an
   * undefined class doesn't invoke autoload.
   */
  SSATmp* checkClass =
    isUnique || (maybeCls && getCurClass() &&
                  getCurClass()->classof(maybeCls))
      ? m_tb->genDefConst(maybeCls)
      : m_tb->gen(LdCachedClass, ssaClassName);

  push(
    haveBit ? m_tb->gen(InstanceOfBitmask,
                        objClass,
                        ssaClassName) :
    isUnique && isNormalClass ? m_tb->gen(ExtendsClass,
                                          objClass,
                                          checkClass) :
    m_tb->gen(InstanceOf,
              objClass,
              checkClass,
              m_tb->genDefConst(maybeCls && !isNormalClass))
  );
  m_tb->genDecRef(src);
}

void HhbcTranslator::emitCastBool() {
  SSATmp* src = popC();
  push(m_tb->genConvToBool(src));
  m_tb->genDecRef(src);
}

void HhbcTranslator::emitCastInt() {
  SSATmp* src = popC();
  push(m_tb->genConvToInt(src));
  m_tb->genDecRef(src);
}

void HhbcTranslator::emitCastDouble() {
  emitInterpOneOrPunt(Type::Dbl, 1);
}

void HhbcTranslator::emitCastString() {
  SSATmp* src = popC();
  Type fromType = src->getType();
  if (fromType == Type::Cell) {
    PUNT(CastString_Cell);
  } else if (fromType == Type::Obj) {
    // call the toString helper on object
    PUNT(CastString_Obj);
  } else {
    // for int to string conversion, this calls a helper that returns
    // a string with ref count of 0.
    pushIncRef(m_tb->genConvToStr(src));
  }
  m_tb->genDecRef(src);
}

void HhbcTranslator::emitCastArray() {
  SSATmp* src = popC();
  pushIncRef(m_tb->genConvToArr(src));
  // TODO: change genConvToArr so that the next line is not needed,
  // but first do Task 2160031.
  m_tb->genDecRef(src);
}

void HhbcTranslator::emitCastObject() {
  emitInterpOneOrPunt(Type::Obj, 1);
}

static
bool isSupportedAGet(SSATmp* classSrc, const StringData* clsName) {
  return (classSrc->isA(Type::Obj) || classSrc->isA(Type::Str) || clsName);
}

void HhbcTranslator::emitAGet(SSATmp* classSrc, const StringData* clsName) {
  if (classSrc->isA(Type::Str)) {
    push(m_tb->gen(LdCls, classSrc, m_tb->genDefConst(getCurClass())));
  } else if (classSrc->isA(Type::Obj)) {
    push(m_tb->gen(LdObjClass, classSrc));
  } else if (clsName) {
    push(m_tb->gen(LdCls,
                   m_tb->genDefConst(clsName),
                   m_tb->genDefConst(getCurClass())));
  } else {
    not_reached();
  }
}

void HhbcTranslator::emitAGetC(const StringData* clsName) {
  if (isSupportedAGet(topC(), clsName)) {
    SSATmp* src = popC();
    emitAGet(src, clsName);
    m_tb->genDecRef(src);
  } else {
    emitInterpOne(Type::Cls, 1);
  }
}

void HhbcTranslator::emitAGetL(int id, const StringData* clsName) {
  SSATmp* src = m_tb->genLdLocAsCell(id, getExitTrace());
  if (isSupportedAGet(src, clsName)) {
    emitAGet(src, clsName);
  } else {
    PUNT(AGetL); // need to teach interpone about local uses
  }
}

SSATmp* HhbcTranslator::unboxPtr(SSATmp* ptr) {
  return m_tb->genUnboxPtr(ptr);
}

void HhbcTranslator::emitBindMem(SSATmp* ptr, SSATmp* src) {
  SSATmp* prevValue = m_tb->genLdMem(ptr, Type::Gen, NULL);
  pushIncRef(src);
  m_tb->genStMem(ptr, src, true);
  m_tb->genDecRef(prevValue);
}

template<class CheckSupportedFun, class EmitLdAddrFun>
void HhbcTranslator::emitBind(const StringData* name,
                              CheckSupportedFun checkSupported,
                              EmitLdAddrFun emitLdAddr) {
  if (!(this->*checkSupported)(name, topC(0)->getType(), 1)) return;
  SSATmp* src = popV();
  emitBindMem((this->*emitLdAddr)(name), src);
}

void HhbcTranslator::emitSetMem(SSATmp* ptr, SSATmp* src) {
  ptr = unboxPtr(ptr);
  SSATmp* prevValue = m_tb->genLdMem(ptr, Type::Cell, nullptr);
  pushIncRef(src);
  m_tb->genStMem(ptr, src, true);
  m_tb->genDecRef(prevValue);
}

template<class CheckSupportedFun, class EmitLdAddrFun>
void HhbcTranslator::emitSet(const StringData* name,
                             CheckSupportedFun checkSupported,
                             EmitLdAddrFun emitLdAddr) {
  if (!(this->*checkSupported)(name, topC(0)->getType(), 1)) return;
  SSATmp* src = popC();
  emitSetMem((this->*emitLdAddr)(name), src);
}

void HhbcTranslator::emitVGetMem(SSATmp* ptr) {
  pushIncRef(m_tb->genLdMem(m_tb->gen(BoxPtr, ptr), Type::BoxedCell, NULL));
}

template<class CheckSupportedFun, class EmitLdAddrFun>
void HhbcTranslator::emitVGet(const StringData* name,
                              CheckSupportedFun checkSupported,
                              EmitLdAddrFun emitLdAddr) {
  if (!(this->*checkSupported)(name, Type::BoxedCell, 0)) return;
  emitVGetMem((this->*emitLdAddr)(name));
}

template<class CheckSupportedFun, class EmitLdAddrFun>
void HhbcTranslator::emitIsset(const StringData* name,
                               CheckSupportedFun checkSupported,
                               EmitLdAddrFun emitLdAddr) {
  if (!(this->*checkSupported)(name, Type::Bool, 0)) return;
  SSATmp* ptr = nullptr;
  SSATmp* result = m_tb->cond(getCurFunc(),
                        [&] (Block* taken) { // branch
                          ptr = (this->*emitLdAddr)(name, taken);
                        },
                        [&] { // Next: property or global is defined
                          return m_tb->gen(IsNTypeMem, Type::Null,
                                          m_tb->genUnboxPtr(ptr));
                        },
                        [&] { // Taken
                          return m_tb->genDefConst(false);
                        }
  );
  push(result);
}

void HhbcTranslator::emitEmptyMem(SSATmp* ptr) {
  SSATmp* ld = m_tb->genLdMem(unboxPtr(ptr), Type::Cell, nullptr);
  push(m_tb->genNot(m_tb->genConvToBool(ld)));
}

template<class CheckSupportedFun, class EmitLdAddrFun>
void HhbcTranslator::emitEmpty(const StringData* name,
                               CheckSupportedFun checkSupported,
                               EmitLdAddrFun emitLdAddr) {
  if (!(this->*checkSupported)(name, Type::Bool, 0)) return;
  SSATmp* ptr = nullptr;
  SSATmp* result = m_tb->cond(getCurFunc(),
                        [&] (Block* taken) {
                          ptr = (this->*emitLdAddr)(name, taken);
                        },
                        [&] { // Next: property or global is defined
                          SSATmp* ld = m_tb->genLdMem(unboxPtr(ptr), Type::Cell,
                                                      nullptr);
                          return m_tb->genNot(m_tb->genConvToBool(ld));
                        },
                        [&] { // Taken
                          return m_tb->genDefConst(true);
                        }
  );
  push(result);
}

void HhbcTranslator::emitBindG(const StringData* gblName) {
  TRACE(3, "%u: BindG\n", m_bcOff);
  emitBind(gblName,
           &HPHP::VM::JIT::HhbcTranslator::checkSupportedGblName,
           &HPHP::VM::JIT::HhbcTranslator::emitLdGblAddrDef);
}

void HhbcTranslator::emitBindS(const StringData* propName) {
  TRACE(3, "%u: BindS\n", m_bcOff);
  emitBind(propName,
           &HPHP::VM::JIT::HhbcTranslator::checkSupportedClsProp,
           &HPHP::VM::JIT::HhbcTranslator::emitLdClsPropAddr);
}

void HhbcTranslator::emitVGetG(const StringData* gblName) {
  TRACE(3, "%u: VGetG\n", m_bcOff);
  emitVGet(gblName,
           &HPHP::VM::JIT::HhbcTranslator::checkSupportedGblName,
           &HPHP::VM::JIT::HhbcTranslator::emitLdGblAddrDef);
}

void HhbcTranslator::emitVGetS(const StringData* propName) {
  TRACE(3, "%u: VGetS\n", m_bcOff);
  emitVGet(propName,
           &HPHP::VM::JIT::HhbcTranslator::checkSupportedClsProp,
           &HPHP::VM::JIT::HhbcTranslator::emitLdClsPropAddr);
}

void HhbcTranslator::emitSetG(const StringData* gblName) {
  emitSet(gblName,
          &HPHP::VM::JIT::HhbcTranslator::checkSupportedGblName,
          &HPHP::VM::JIT::HhbcTranslator::emitLdGblAddrDef);
}

void HhbcTranslator::emitSetS(const StringData* propName) {
  emitSet(propName,
          &HPHP::VM::JIT::HhbcTranslator::checkSupportedClsProp,
          &HPHP::VM::JIT::HhbcTranslator::emitLdClsPropAddr);
}

static Type getResultType(Type resultType, bool isInferedType) {
  assert(!isInferedType || resultType.isKnownUnboxedDataType());
  if (resultType.equals(Type::None)) {
    // result type neither predicted nor inferred
    return Type::Cell;
  }
  assert(resultType.isKnownUnboxedDataType());
  return resultType;
}

template<class CheckSupportedFun, class EmitLdAddrFun>
void HhbcTranslator::emitCGet(const StringData* name,
                              Type resultType,
                              bool isInferedType,
                              bool exitOnFailure,
                              CheckSupportedFun checkSupported,
                              EmitLdAddrFun emitLdAddr) {
  resultType = getResultType(resultType, isInferedType);
  if (!(this->*checkSupported)(name, resultType, 0)) return;
  Trace* exit = (isInferedType || resultType.equals(Type::Cell))
                ? nullptr : getExitSlowTrace();
  SSATmp* ptr = (this->*emitLdAddr)(name,
                                    exitOnFailure
                                      ? m_tb->getFirstBlock(getExitSlowTrace())
                                      : nullptr);
  if (!isInferedType) ptr = unboxPtr(ptr);
  pushIncRef(m_tb->genLdMem(ptr, resultType, exit));
}

void HhbcTranslator::emitCGetG(const StringData* gblName,
                               Type resultType,
                               bool isInferedType) {
  TRACE(3, "%u: CGetG\n", m_bcOff);
  emitCGet(gblName, resultType, isInferedType, true,
           &HPHP::VM::JIT::HhbcTranslator::checkSupportedGblName,
           &HPHP::VM::JIT::HhbcTranslator::emitLdGblAddr);
}

void HhbcTranslator::emitCGetS(const StringData* propName,
                               Type resultType,
                               bool isInferedType) {
  TRACE(3, "%u: CGetS\n", m_bcOff);
  emitCGet(propName, resultType, isInferedType, false,
           &HPHP::VM::JIT::HhbcTranslator::checkSupportedClsProp,
           &HPHP::VM::JIT::HhbcTranslator::emitLdClsPropAddrOrExit);
}

void HhbcTranslator::emitBinaryArith(Opcode opc) {
  bool isBitOp = (opc == OpAnd || opc == OpOr || opc == OpXor);
  Type type1 = topC(0)->getType();
  Type type2 = topC(1)->getType();
  if (isSupportedBinaryArith(opc, type1, type2)) {
    SSATmp* tr = popC();
    SSATmp* tl = popC();
    push(m_tb->gen(opc, Type::binArithResultType(type1, type2), tl, tr));
  } else if (isBitOp && (type1 == Type::Obj || type2 == Type::Obj)) {
    // raise fatal
    emitInterpOne(Type::Cell, 2);
  } else {
    Type type = Type::Int;
    if (isBitOp) {
      if (type1.isString() && type2.isString()) {
        type = Type::Str;
      } else if ((type1.needsReg() && (type2.needsReg() || type2.isString()))
                 || (type2.needsReg() && type1.isString())) {
        // both types might be strings, but can't tell
        type = Type::Cell;
      } else {
        type = Type::Int;
      }
    } else {
      // either an int or a dbl, but can't tell
      type = Type::Cell;
    }
    emitInterpOne(type, 2);
  }
}

void HhbcTranslator::emitNot() {
  TRACE(3, "%u: Not\n", m_bcOff);
  SSATmp* src = popC();
  push(m_tb->genNot(m_tb->genConvToBool(src)));
  m_tb->genDecRef(src);
}

void HhbcTranslator::emitAdd() {
  TRACE(3, "%u: Add\n", m_bcOff);
  emitBinaryArith(OpAdd);
}
void HhbcTranslator::emitSub() {
  TRACE(3, "%u: Sub\n", m_bcOff);
  emitBinaryArith(OpSub);
}
void HhbcTranslator::emitMul() {
  TRACE(3, "%u: Mul\n", m_bcOff);
  emitBinaryArith(OpMul);
}
void HhbcTranslator::emitBitNot() {
  TRACE(3, "%u: BitNot\n", m_bcOff);
  Type srcType = topC()->getType();
  if (srcType == Type::Int) {
    SSATmp* src = popC();
    SSATmp* ones = m_tb->genDefConst<int64_t>(~0);
    push(m_tb->genXor(src, ones));
  } else if (srcType.subtypeOf(Type::Null | Type::Bool | Type::Arr | Type::Obj)) {
    // raise fatal
    emitInterpOne(Type::Cell, 1);
  } else {
    Type resultType = Type::Int;
    if (srcType.isString()) {
      resultType = Type::Str;
    } else if (srcType.needsReg()) {
      resultType = Type::Cell;
    }
    emitInterpOne(resultType, 1);
  }
}
void HhbcTranslator::emitBitAnd() {
  TRACE(3, "%u: BitAnd\n", m_bcOff);
  emitBinaryArith(OpAnd);
}
void HhbcTranslator::emitBitOr() {
  TRACE(3, "%u: BitOr\n", m_bcOff);
  emitBinaryArith(OpOr);
}
void HhbcTranslator::emitBitXor() {
  TRACE(3, "%u: BitXor\n", m_bcOff);
  emitBinaryArith(OpXor);
}
void HhbcTranslator::emitXor() {
  TRACE(3, "%u: Xor\n", m_bcOff);
  SSATmp* btr = popC();
  SSATmp* btl = popC();
  SSATmp* tr = m_tb->genConvToBool(btr);
  SSATmp* tl = m_tb->genConvToBool(btl);
  push(m_tb->genConvToBool(m_tb->genXor(tl, tr)));
  m_tb->genDecRef(btl);
  m_tb->genDecRef(btr);
}

void HhbcTranslator::emitInterpOne(Type type,
                                   int numDiscard, /* = 0 */
                                   Trace* target /* = NULL */) {
  spillStack();
  // discard the top elements of the stack
  discard(numDiscard);
  m_tb->genInterpOne(m_bcOff, m_stackDeficit, type, target);
  m_stackDeficit = 0;
}

void HhbcTranslator::emitInterpOneOrPunt(Type type,
                                         int numDiscard, /* = 0 */
                                         Trace* target /* = NULL */) {
  if (RuntimeOption::EvalIRPuntDontInterp) {
    PUNT(PuntDontInterp);
  } else {
    emitInterpOne(type, numDiscard, target);
  }
}

Trace* HhbcTranslator::getGuardExit() {
  assert(m_bcOff == -1 || m_bcOff == m_startBcOff);
  // stack better be empty since we're at the start of the trace
  assert((m_evalStack.numCells() - m_stackDeficit) == 0);
  return m_exitGuardFailureTrace;
}

/*
 * Get SSATmps representing all the information on the virtual eval
 * stack in preparation for a spill or exit trace.
 *
 * Doesn't actually remove these values from the eval stack.
 */
std::vector<SSATmp*> HhbcTranslator::getSpillValues() const {
  std::vector<SSATmp*> ret;
  ret.reserve(m_evalStack.size());
  for (int i = 0; i < m_evalStack.size(); ++i) {
    SSATmp* elem = m_evalStack.top(i);

    ret.push_back(elem);
    if (elem->getType() == Type::ActRec) {
      /*
       * For register allocation purposes, the SpillStack instruction
       * should count as a use of each of the SSATmps coming into the
       * a DefActRec.
       */
      auto* inst = elem->getInstruction();
      assert(inst->getNumSrcs() == 5);
      ret.push_back(inst->getSrc(1)); // func
      ret.push_back(inst->getSrc(2)); // objOrCls
      ret.push_back(inst->getSrc(3)); // numArgs
      ret.push_back(inst->getSrc(4)); // invName
    }
  }
  return ret;
}

/*
 * Generates an exit trace which will continue the VM execution at the given
 * nextByteCode (defaults to the current m_bcOff) without using HHIR.
 * This should be used in situations that HHIR cannot handle -- ideally only in
 * slow paths.
 */
Trace* HhbcTranslator::getExitSlowTrace() {
  std::vector<SSATmp*> stackValues = getSpillValues();
  return m_tb->getExitSlowTrace(m_bcOff,
                                m_stackDeficit,
                                stackValues.size(),
                                stackValues.size() ? &stackValues[0] : 0);
}

/*
 * Generates an exit trace for the given targetBcOff
 * (defaults to the current offset).
 * The exit trace returned will be linked to a translation starting at
 * targetBcOff, which will be a retranslation of the same tracelet if this
 * exit is taken before executing any bytecode instruction of the current
 * tracelet.
 */
Trace* HhbcTranslator::getExitTrace(Offset targetBcOff /* = -1 */) {
  if (targetBcOff == -1) {
    targetBcOff = m_bcOff != -1 ? m_bcOff : m_startBcOff;
  }
  if (targetBcOff == m_startBcOff) {
    return m_exitGuardFailureTrace;
  }

  std::vector<SSATmp*> stackValues = getSpillValues();
  return m_tb->genExitTrace(targetBcOff,
                           m_stackDeficit,
                           stackValues.size(),
                           stackValues.size() ? &stackValues[0] : nullptr,
                           TraceExitType::Normal);
}

/*
 * Generates a trace exit that can be the target of a conditional
 * control flow instruction at the current bytecode offset.
 */
Trace* HhbcTranslator::getExitTrace(uint32_t targetBcOff, uint32_t notTakenBcOff) {
  std::vector<SSATmp*> stackValues = getSpillValues();
  return m_tb->genExitTrace(targetBcOff,
                           m_stackDeficit,
                           stackValues.size(),
                           stackValues.size() ? &stackValues[0] : nullptr,
                           TraceExitType::NormalCc,
                           notTakenBcOff);
}

SSATmp* HhbcTranslator::spillStack() {
  std::vector<SSATmp*> stackValues = getSpillValues();
  m_evalStack.clear();
  SSATmp* sp = m_tb->genSpillStack(m_stackDeficit,
                                   stackValues.size(),
                                   stackValues.size() ? &stackValues[0] : 0);
  m_stackDeficit = 0;
  return sp;
}

SSATmp* HhbcTranslator::loadStackAddr(int32_t offset) {
  // You're almost certainly doing it wrong if you want to get the address of a
  // stack cell that's in m_evalStack.
  assert(offset >= (int32_t)m_evalStack.numCells());
  return m_tb->genLdStackAddr(
    offset + m_stackDeficit - m_evalStack.numCells());
}

//
// This is a wrapper to TraceBuilder::genLdLoc() that also emits the
// RaiseUninitLoc if the local is uninitialized
//
SSATmp* HhbcTranslator::emitLdLocWarn(uint32_t id,
                                      Trace* target) {
  SSATmp* locVal = m_tb->genLdLocAsCell(id, target);

  if (locVal->getType().subtypeOf(Type::Uninit)) {
    spillStack();
    m_tb->genRaiseUninitLoc(id);
    return m_tb->genDefInitNull();
  }

  return locVal;
}

void HhbcTranslator::end(int nextPc) {
  if (m_hasExit) return;

  if (nextPc >= getCurFunc()->past()) {
    // We have fallen off the end of the func's bytecodes. This happens
    // when the function's bytecodes end with an unconditional
    // backwards jump so that nextPc is out of bounds and causes an
    // assertion failure in unit.cpp. The common case for this comes
    // from the default value funclets, which are placed after the end
    // of the function, with an unconditional branch back to the start
    // of the function. So you should see this in any function with
    // default params.
    return;
  }
  setBcOff(nextPc, true);
  spillStack();
  m_tb->genTraceEnd(nextPc);
}

}}} // namespace HPHP::VM::JIT
